/**
  *  \file t_utf8.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/utf8.h"

#include <cstdlib>
#include "tests.h"

/** Test that encoding and then decoding produces identical results. */
void
TestUtf8::testEncDec()
{
    for (uint32_t i = 0; i < 0x10000; ++i) {
        string_t tmp;
        appendUtf8(tmp, i);
        Utf8Reader rdr(tmp);
        TS_ASSERT(rdr.hasMore());
        uint32_t g = rdr.get();
        TS_ASSERT_EQUALS(g, i);
        TS_ASSERT(!rdr.hasMore());
    }

    // Test same thing for error characters: those are now preserved.
    {
        string_t e;
        appendUtf8(e, Utf8Reader::makeErrorCharacter(0xC2));
        appendUtf8(e, Utf8Reader::makeErrorCharacter(0x80));
        Utf8Reader rdr(e);
        TS_ASSERT(rdr.hasMore());
        uint32_t g = rdr.get();
        TS_ASSERT_EQUALS(g, Utf8Reader::makeErrorCharacter(0xC2));
        TS_ASSERT(rdr.hasMore());
        g = rdr.get();
        TS_ASSERT_EQUALS(g, Utf8Reader::makeErrorCharacter(0x80));
        TS_ASSERT(!rdr.hasMore());
    }
}

/** Test that decoding and then encoding produces identical results. */
void
TestUtf8::testDecEnc()
{
    for (int i = 0; i < 256; ++i) {
        for (int j = 0; j < 256; ++j) {
            for (int k = 0; k < 256; k += 3) { /* step 3 to speed this up a little */
                char tmp[3] = { i, j, k };
    
                uint32_t u[10];
                int n = 0;
                Utf8Reader rdr(tmp, 3);
                while (rdr.hasMore())
                    u[n++] = rdr.get();
                TS_ASSERT_EQUALS(n, Utf8Reader(tmp, 3).count());

                // This no longer holds, just check that it doesn't crash above
                // string_t tmp2; tmp2.reserve(3);
                // for (int m = 0; m < n; ++m)
                //     appendUtf8(tmp2, u[m]);
                // 
                // TS_ASSERT_EQUALS(tmp2.size(), 3);
                // TS_ASSERT_EQUALS(tmp2.compare(0, 3, tmp, 3), 0);
            }
        }
    }
}

/** Test decoding random strings. */
void
TestUtf8::testFuzz()
{
    static const int NCHAR = 100;
    std::srand(42);
    for (int i = 0; i < 1000; ++i) {
        string_t tmp;
        uint32_t u[NCHAR];
        for (int j = 0; j < NCHAR; ++j)
            tmp += char(rand() & 0xFF);

        int n = 0;
        Utf8Reader rdr(tmp);
        while (rdr.hasMore()) {
            TS_ASSERT_LESS_THAN_EQUALS(n, NCHAR);
            u[n++] = rdr.get();
        }

        // This no longer holds, so we just check it doesn't crash above
        // string_t tmp2; tmp.reserve(NCHAR);
        // for (int m = 0; m < n; ++m)
        //     appendUtf8(tmp2, u[m]);
        //
        // TS_ASSERT_EQUALS(tmp, tmp2);
    }
}

/** Test lengthUtf8(). */
void
TestUtf8::testLength()
{
    TS_ASSERT_EQUALS(lengthUtf8(""),             0);
    TS_ASSERT_EQUALS(lengthUtf8("123"),          3);    /* ASCII ASCII ASCII */
    TS_ASSERT_EQUALS(lengthUtf8("\xc2""23"),     3);    /* ERROR ASCII ASCII */
    TS_ASSERT_EQUALS(lengthUtf8("\xc2\x80""3"),  2);    /* <--RUNE-->  ASCII */
    TS_ASSERT_EQUALS(lengthUtf8("\xc2\x80\xc2"), 2);    /* <--RUNE-->  ERROR */
}

/** Test substrUtf8(). */
void
TestUtf8::testSubstr()
{
    TS_ASSERT_EQUALS(substrUtf8("", 10, 10), string_t(""));
}
